use sqlx::SqlitePool;

use crate::types::uuid::Uuid;

use super::{sessions::Session, users::User};

pub struct Device {
    uuid: Uuid,
    user_uuid: Uuid,
    device_id: String,
    display_name: String,
}

impl Device {
    pub fn new(user: &User, device_id: &str, display_name: &str) -> anyhow::Result<Self> {
        Ok(Self {
            uuid: uuid::Uuid::new_v4().into(),
            user_uuid: user.uuid().clone(),
            device_id: device_id.to_owned(),
            display_name: display_name.to_owned(),
        })
    }

    pub async fn create(&self, conn: &SqlitePool) -> anyhow::Result<Self> {
        Ok(sqlx::query_as!(
            Self,
            "insert into devices(uuid, user_uuid, device_id, display_name)
            values(?, ?, ?, ?)
            returning uuid as 'uuid: Uuid', user_uuid as 'user_uuid: Uuid', device_id, display_name",
            self.uuid,
            self.user_uuid,
            self.device_id,
            self.display_name)
            .fetch_one(conn).await?
        )
    }

    pub async fn find_for_user(
        conn: &SqlitePool,
        user: &User,
        device_id: &str,
    ) -> anyhow::Result<Self> {
        let user_uuid = user.uuid();
        Ok(sqlx::query_as!(
            Self,
            "select uuid as 'uuid: Uuid', user_uuid as 'user_uuid: Uuid', device_id, display_name from devices where user_uuid = ?",
            user_uuid)
            .fetch_one(conn).await?
        )
    }

    /// Get the device's id.
    #[must_use]
    pub fn uuid(&self) -> &Uuid {
        &self.uuid
    }

    /// Get the device's user id.
    #[must_use]
    pub fn user_uuid(&self) -> &Uuid {
        &self.user_uuid
    }

    /// Get a reference to the device's device id.
    #[must_use]
    pub fn device_id(&self) -> &str {
        self.device_id.as_ref()
    }

    /// Get a reference to the device's display name.
    #[must_use]
    pub fn display_name(&self) -> &str {
        self.display_name.as_ref()
    }
}
