use rand::{thread_rng, Rng, distributions::Alphanumeric};
use sqlx::SqlitePool;

use crate::types::uuid::Uuid;

use super::devices::Device;

pub struct Session {
    uuid: Uuid,
    device_uuid: Uuid,
    key: String,
}

impl Session {
    pub fn new(device: &Device) -> anyhow::Result<Self> {
        let mut rng = thread_rng();
        let key: String = (0..32).map(|_| rng.sample(Alphanumeric) as char).collect();

        Ok(Self {
            uuid: uuid::Uuid::new_v4().into(),
            device_uuid: device.uuid().clone(),
            key,
        })
    }

    pub async fn create(&self, conn: &SqlitePool) -> anyhow::Result<Self> {
        Ok(sqlx::query_as!(
            Self,
            "insert into sessions(uuid, device_uuid, key)
            values(?, ?, ?)
            returning uuid as 'uuid: Uuid', device_uuid as 'device_uuid: Uuid', key",
            self.uuid,
            self.device_uuid,
            self.key
        )
        .fetch_one(conn)
        .await?)
    }

    /// Get the session's id.
    #[must_use]
    pub fn uuid(&self) -> &Uuid {
        &self.uuid
    }

    /// Get the session's device id.
    #[must_use]
    pub fn device_uuid(&self) -> &Uuid {
        &self.device_uuid
    }

    /// Get a reference to the session's value.
    #[must_use]
    pub fn key(&self) -> &str {
        self.key.as_ref()
    }
}
