use crate::types::uuid::Uuid;
use argon2::{password_hash::SaltString, Argon2, PasswordHasher, PasswordHash, PasswordVerifier};
use rand::rngs::OsRng;
use sqlx::{encode::IsNull, sqlite::SqliteTypeInfo, FromRow, Sqlite, SqlitePool};

use crate::types::user_id::UserId;

pub struct User {
    uuid: Uuid,
    user_id: UserId,
    display_name: String,
    password: String,
}

impl User {
    pub fn new(user_id: &UserId, display_name: &str, password: &str) -> anyhow::Result<Self> {
        let argon2 = Argon2::default();
        let salt = SaltString::generate(OsRng);
        let password = argon2
            .hash_password(password.as_bytes(), &salt)?
            .to_string();

        Ok(Self {
            uuid: uuid::Uuid::new_v4().into(),
            user_id: user_id.clone(),
            display_name: display_name.to_owned(),
            password,
        })
    }

    pub async fn exists(conn: &SqlitePool, user_id: &UserId) -> anyhow::Result<bool> {
        Ok(
            sqlx::query!("select user_id from users where user_id = ?", user_id)
                .fetch_optional(conn)
                .await?
                .is_some(),
        )
    }

    pub async fn create(&self, conn: &SqlitePool) -> anyhow::Result<Self> {
        Ok(sqlx::query_as!(
            Self,
            "insert into users(uuid, user_id, display_name, password)
            values (?, ?, ?, ?)
            returning uuid as 'uuid: Uuid', user_id as 'user_id: UserId', display_name, password",
            self.uuid,
            self.user_id,
            self.display_name,
            self.password
        )
        .fetch_one(conn)
        .await?)
    }

    pub async fn update(&self, conn: &SqlitePool) -> anyhow::Result<Self> {
        Ok(sqlx::query_as!(
            Self,
            "update users set uuid = ?, user_id = ?, display_name = ?, password = ?
            where uuid = ?
            returning uuid as 'uuid: Uuid', user_id as 'user_id: UserId', display_name, password",
            self.uuid,
            self.user_id,
            self.display_name,
            self.password,
            self.uuid
        )
        .fetch_one(conn)
        .await?)
    }

    pub async fn find_by_user_id(conn: &SqlitePool, user_id: &UserId) -> anyhow::Result<Self> {
        Ok(sqlx::query_as!(
            Self,
            "select uuid as 'uuid: Uuid', user_id as 'user_id: UserId', display_name, password
             from users where user_id = ?",
            user_id
        )
        .fetch_one(conn)
        .await?)
    }

    pub fn password_correct(&self, password: &str) -> anyhow::Result<bool> {
        let password_hash = PasswordHash::new(self.password())?;

        Ok(Argon2::default().verify_password(password.as_bytes(), &password_hash).is_ok())
    }

    /// Get the user's id.
    #[must_use]
    pub fn uuid(&self) -> &Uuid {
        &self.uuid
    }

    /// Get a reference to the user's user id.
    #[must_use]
    pub fn user_id(&self) -> &UserId {
        &self.user_id
    }

    /// Get a reference to the user's password.
    #[must_use]
    pub fn password(&self) -> &str {
        self.password.as_ref()
    }
}
